# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .squared_correlation import SquaredCorrelation  # noqa: F401
from .squared_correlation import SquaredCorrelation as IntegralVarianceReduction  # noqa: F401
from .mutual_information import MutualInformation  # noqa: F401
from .uncertainty_sampling import UncertaintySampling  # noqa: F401
