# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0


from .linear_model import GPyLinearMultiFidelityModel  # noqa: F401
from .non_linear_multi_fidelity_model import NonLinearMultiFidelityModel  # noqa: F401
