from .borehole import multi_fidelity_borehole_function  # noqa: F401
from .branin import multi_fidelity_branin_function  # noqa: F401
from .currin import multi_fidelity_currin_function  # noqa: F401
from .hartmann import multi_fidelity_hartmann_3d  # noqa: F401
from .park import multi_fidelity_park_function  # noqa: F401
